#include "linker.h"
#include "../utils/name2type.h"
#include "../../__common/binaries.h"
#include "../../__common/objects_types.h"

#include "linker_context.h"
#include "standard_functions.h"

#include <stdexcept>

namespace linker
{
	class global_context : public context
	{
	public:

		global_context(name2index& links, 
			name2expr& exprs, 
			pointer_vector<object>& objects);

		expression::base* getExpr(const std::string& name) const;
		virtual binary::link getLink(const std::string& name) const;
		virtual binary::link add(object* obj);

	private:

		const name2index& m_links;
		const name2expr& m_exprs;
		pointer_vector<object>& m_objects;
	};
};

////////////////////////////////////////
//
// Program
//
////////////////////////////////////////

linker::program::program(syntacter::program& p)
{
	name2index links;
	name2expr exprs;

	// make name-to-index table. We do it before
	// all, because any definition can use any other
	{for (size_t i = 0; i < p.defs().size(); i++)
	{
		links.add(p.defs().name(i), (int)i);
		exprs.add(p.defs().name(i), p.defs().exp(i));
	}}

	// reserve first N items in object table 
	// for named objects
	m_objects.resize(p.defs().size());

	global_context gc(links, exprs, m_objects);

	{for (size_t i = 0; i < p.defs().size(); i++)
	{
		m_objects.set(i, object::create(p.defs().name(i), p.defs().exp(i)->link(gc)));
	}}

	m_objects.push_back(object::create("", p.exp()->link(gc)));
}

linker::program::~program()
{
}

size_t linker::program::size() const
{
	return m_objects.size();
}

const linker::object& linker::program::operator[](size_t i) const
{
	if (i < m_objects.size())
	{
		return *m_objects[i];
	}

	throw std::runtime_error("linker::program::operator[]: too big index");
}

//////////////////////////////////
//
// Global context
//
//////////////////////////////////

linker::global_context::global_context(
				name2index& links, 
				name2expr& exprs, 
				pointer_vector<object>& objects)
					: m_links(links), m_exprs(exprs), m_objects(objects)
{
}

expression::base* linker::global_context::getExpr(const std::string& name) const
{
	expression::base* e = m_exprs[name];

	if (e != 0)
	{
		return e;
	}

	e = linker::getStandardExprs()[name];

	if (e != 0)
	{
		return e;
	}

	throw std::runtime_error("linker::global_context::getExpr: can not find name");
}

binary::link linker::global_context::getLink(const std::string& name) const
{
	binary::link l;

	int index = m_links[name];

	if (index >= 0)
	{
		l.l.table = binary::link::program;
		l.l.item = index;
		return l;
	}

	expression::base* exp = linker::getStandardExprs()[name];

	if (exp != 0)
	{
		return exp->link(*const_cast<global_context*>(this));
	}

	throw std::runtime_error("linker::global_context::getLink: can not find name");
}

binary::link linker::global_context::add(object* obj)
{
	m_objects.push_back(obj);

	binary::link l;
	l.l.table = binary::link::program;
	l.l.item = (unsigned int)m_objects.size() - 1;
	return l;
}
