#include "object_base.h"

#include "object_literals.h"
#include "object_code.h"
#include "object_lambda.h"

#include "../tasks/pool.h"
#include "../../__common/objects_types.h"

//
// Common create
//

object::proxy object::create(const loader::record& rec, bool separable, const context::base* c)
{
	switch (rec.type())
	{
		// literals
		case binary::types::numeric:    return numeric::create(rec.data(), rec.size());
		case binary::types::string:     return string::create(rec.data(), rec.size());
		case binary::types::symbol:     return symbol::create(rec.data(), rec.size());
		case binary::types::character:  return character::create(rec.data(), rec.size());
		case binary::types::boolean:    return boolean::create(rec.data(), rec.size());
		case binary::types::pair:       return pair::create(c, rec.data(), rec.size());
		case binary::types::empty_list: return empty_list::create(rec.data(), rec.size());

		// code
		case binary::types::call: return call::create(c, rec, separable);
		case binary::types::if_:  return if_::create(c, rec, separable);
		case binary::types::cond: break;
		case binary::types::or_:  break;
		case binary::types::and_: break;

		// lambda
		case binary::types::lambda: return lambda::create(c, rec.data(), rec.size());
	}

	if (rec.type() == binary::types::link)
	{
		// link
		if (rec.size() != sizeof(binary::link))
		{
			throw std::runtime_error("object::link::create: wrong size!");
		}

		const binary::link* pl = (const binary::link*)rec.data();
		return c->get(*pl, separable);		
	}

	throw std::runtime_error("object::create: unknown object type");
}

///////////////////////////////
//
// Abstract object
//
///////////////////////////////

object::base::base(int t) : m_type(t), m_c(0), m_simple(true)
{
	m_recursive = false;
}

object::base::base(int t, const context::base* c, int est, bool simple) 
			: m_type(t), m_c(c), m_simple(simple)
{
	if ((m_c.get() == 0) && !m_simple)
	{
		throw std::runtime_error("object::base::base: non simple object with null context");
	}

	if ((est >= (1 << 16)) || (est < 0))
	{
		m_recursive = true;
	}
	else
	{
		m_recursive = false;
	}
}

object::base::~base()
{
}


bool object::base::equiv(const base* o) const
{
	return false;
}

bool object::base::isTrue() const
{
	return true;
}

object::proxy object::base::evaluate(bool separable) const
{
	// Default simple evaluation for simple objects
	//proxy p(this);
	//return p;
	throw std::runtime_error("object::base::evaluate: object already simple!");
}

object::proxy object::base::apply(bool separable, size_t argc, const proxy* args) const
{
	throw std::runtime_error("object::base::apply: it is not a function!");
}

/////////////////////////////////////////
//
// Proxy
//
//////////////////////////////////////////

object::proxy::proxy() : m_p(0), m_separable(false), m_id(0)
{
}

object::proxy::proxy(const base* p) : m_p(p), m_separable(false), m_id(0)
{
	if (!m_p->isSimple())
	{
		throw std::runtime_error("object::proxy::proxy: non simple object with simple constructor");
	}			
}

object::proxy::proxy(const base* p, bool separable) : m_p(p), m_separable(separable), m_id(0)
{
	if (m_separable && p->isRecursive())
	{
		if (tasks::getPool().size() > 0)
		{
			// Yes! Create task and push it into pool
			m_id = tasks::getPool().push_task(*this);
		}
	}

	/*
	// this code make realisation NOT lazy
	// use it just for test
	proxy tmp = simplify();
	m_p = tmp.m_p;
	*/
}

object::proxy object::proxy::simplify() const
{
	proxy p = *this;

	for (;;)
	{
		if (p->isSimple()) 
		{
			break;
		}
		
		if (p.m_id) 
		{
			p = tasks::getPool().getResult(p.m_id);
			break;
		}

		p = p->evaluate(m_separable);
	}

	return p;
}
