#include "object_literals.h"

#include <stdlib.h>
#include <math.h>
#include <stdexcept>

#ifndef _MSC_VER
#define _CVTBUFSIZE 100
#endif

// for simplify definitions
using object::proxy;
using object::base;

//////////////////////////	
//
// Numeric object
//
//////////////////////////

//
// Creation
//

proxy object::numeric::create(double v)
{
	return proxy(new numeric(v));
}

proxy object::numeric::create(const char* v, size_t size)
{
	if (size != sizeof(binary::numeric))
	{
		throw std::runtime_error("object::numeric::create: wrong size!");
	}

	const binary::numeric* p = (const binary::numeric*)v;
	return proxy(new numeric(p->value));
}

object::numeric::numeric(double v) : base(object::types::numeric), m_value(v)
{
}

//
// Value
//

double object::numeric::value() const
{
	return m_value;
}

//
// Overrides
//

bool object::numeric::equiv(const base* o) const
{
	if (o->type() != type())
	{
		return false;
	}

	const numeric* n = (const numeric*)o;
	return (m_value == n->value());
}

std::string object::numeric::toString() const
{
	char buff[_CVTBUFSIZE];

	if ((m_value == floor(m_value)) && (m_value < 0x7fffffffl))
	{
#if defined(_MSC_VER) && (_MSC_VER >= 1400)
		_itoa_s((int)m_value, buff, sizeof(buff), 10);
		return buff;
#elif defined(_MSC_VER)
		return itoa((int)m_value, buff, 10);
#else
		return gcvt(m_value, 20, buff);
#endif
	}

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	_gcvt_s(buff, sizeof(buff), m_value, 20);
	return buff;
#else
	return gcvt(m_value, 20, buff);
#endif
}

/////////////////////////////////
//
// Boolean object
//
/////////////////////////////////

//
// creation
//

proxy object::boolean::create(bool v)
{
	return proxy(new boolean(v));
}

proxy object::boolean::create(const char* v, size_t size)
{
	if (size != sizeof(binary::boolean))
	{
		throw std::runtime_error("object::boolean::create: wrong size!");
	}

	const binary::boolean* p = (const binary::boolean*)v;
	return proxy(new boolean(p->value != 0));
}

object::boolean::boolean(bool v) : base(object::types::boolean), m_value(v)
{
}

//
// Value
//

bool object::boolean::value() const
{
	return m_value;
}

//
// Overrides
//

bool object::boolean::equiv(const base* o) const
{
	if (o->type() != type())
	{
		return false;
	}

	const boolean* n = (const boolean*)o;
	return (m_value == n->value());
}

bool object::boolean::isTrue() const
{
	return m_value;
}

std::string object::boolean::toString() const
{
	if (m_value)
	{
		return "#t";
	}

	return "#f";
}

/////////////////////////////
//
// Character object
//
/////////////////////////////

//
// Creation
//

proxy object::character::create(char v)
{
	return proxy(new character(v));
}

proxy object::character::create(const char* v, size_t size)
{
	if (size != sizeof(binary::character))
	{
		throw std::runtime_error("object::character::create: wrong size!");
	}

	const binary::character* p = (const binary::character*)v;
	return proxy(new character(p->value));
}

object::character::character(char v) : base(object::types::character), m_value(v)
{
}

//
// Value
//

char object::character::value() const
{
	return m_value;
}

//
// Overrides
//

bool object::character::equiv(const base* o) const
{
	if (o->type() != type())
	{
		return false;
	}

	const character* n = (const character*)o;
	return (m_value == n->value());
}


std::string object::character::toString() const
{
	char buff[2] = {0};
	buff[0] = m_value;
	return buff;
}

//////////////////////////////////////////////
//
// String object
//
//////////////////////////////////////////////

//
// Creation
//

proxy object::string::create(const std::string& v)
{
	return proxy(new string(v.c_str(), v.length()));
}

proxy object::string::create(const char* v, size_t size)
{
	return proxy(new string(v, size));	
}

object::string::string(const char* v, size_t size) : base(object::types::string)
{
	m_value.assign(v, v + size);
}

//
// Value
//

const std::string& object::string::value() const
{
	return m_value;
}

//
// Overrides
//

bool object::string::equiv(const base* o) const
{
	if (o->type() != type())
	{
		return false;
	}

	const string* n = (const string*)o;
	return (m_value == n->value());
}


std::string object::string::toString() const
{
	std::string str = "\"";
	str += m_value;
	str += "\"";
	return str;
}

//////////////////////////////////////////
//
// Symbol object
//
///////////////////////////////////////////

//
// Creation
//

proxy object::symbol::create(const std::string& v)
{
	return proxy(new symbol(v.c_str(), v.length()));
}

proxy object::symbol::create(const char* v, size_t size)
{
	if (size < 1)
	{
		throw std::runtime_error("object::symbol::create: wrong size!");
	}

	return proxy(new symbol(v, size));	
}

object::symbol::symbol(const char* v, size_t size) : base(object::types::symbol)
{
	m_value.assign(v, v + size);
}

//
// Overrides
//

bool object::symbol::equiv(const base* o) const
{
	if (o->type() != type())
	{
		return false;
	}

	const symbol* n = (const symbol*)o;
	return (m_value == n->m_value);
}

std::string object::symbol::toString() const
{
	return m_value;
}

/////////////////////////////////////////
//
// Pair object
//
/////////////////////////////////////////

//
// Creation
// 

proxy object::pair::create(const context::base* c, const char* v, size_t size)
{
	if (size != sizeof(binary::pair))
	{
		throw std::runtime_error("object::pair::create: wrong size!");
	}

	const binary::pair* p = (const binary::pair*)v;

	return proxy(new pair(c, p->car, p->cdr), false);
}

proxy object::pair::cons(proxy head, proxy tail)
{
	return proxy(new pair(head, tail));
}

proxy object::pair::list(size_t c, const proxy* item)
{
	// tail of list
	proxy p = empty_list::create();

	// go from last to first argument
	for (int i = (int)c - 1; i >= 0; i--)
	{
		p = cons(item[i], p);
	}

	return p;
}

object::pair::pair(const context::base* c, binary::link head, binary::link tail) 
				: base(object::types::pair, c, 1)
{
	m_carLink = head;
	m_cdrLink = tail;
}

object::pair::pair(proxy head, proxy tail) : base(object::types::pair)
{
	m_car = head;
	m_cdr = tail;
}

//
// Useful functions
//

proxy object::pair::car() const
{
	return m_car;
}

proxy object::pair::cdr() const
{
	return m_cdr;
}

bool object::pair::isList() const
{
	m_cdr = m_cdr.simplify();

	if (m_cdr->isNull())
	{
		return true;
	}

	if (m_cdr->isPair())
	{
		const pair* p = m_cdr.get<pair, object::types::pair>();
		return p->isList();
	}

	return false;
}

//
// Overrides
//

proxy object::pair::evaluate(bool separable) const
{
	if (!m_simple)
	{
		m_car = m_c->get(m_carLink, false);
		m_cdr = m_c->get(m_cdrLink, false);
		m_simple = true;
		m_c = 0;
	}

	proxy p(this);
	return p;
}

std::string object::pair::toString() const
{
	m_car = m_car.simplify();
	m_cdr = m_cdr.simplify();

	std::string str;
	str = "(";
	str += m_car->toString();

	if (isList())
	{
		str += " ";

		const pair* p = m_cdr.get<pair, object::types::pair>();
		str += p->toListString();
	}
	else
	{
		str += " . ";
		str += m_cdr->toString();
		str += ")";
	}

	return str;
}

std::string object::pair::toListString() const 
{
	m_car = m_car.simplify();
	m_cdr = m_cdr.simplify();

	std::string str;
	str += m_car->toString();

	if (m_cdr->isNull())
	{
		str += ")";
	}
	else
	{
		str += " ";
		const pair* p = m_cdr.get<pair, object::types::pair>();
		str += p->toListString();
	}

	return str;
}

////////////////////////////////////
//
// Empty_list object
//
////////////////////////////////////

//
// Creation
//

proxy object::empty_list::create(const char* v, size_t size)
{
	if (size != 0)
	{
		throw std::runtime_error("object::empty_list::create: wrong size!");
	}

	return proxy(new empty_list());
}

proxy object::empty_list::create()
{
	return proxy(new empty_list());
}

object::empty_list::empty_list() : base(object::types::empty_list)
{
}

//
// Overrides
//

bool object::empty_list::equiv(const base* o) const
{
	return (o->type() == type());
}

std::string object::empty_list::toString() const
{
	return "()";
}

