#include "object_stdfunc.h"
#include "object_literals.h"
#include "../utils/myvector.h"

#include <stdexcept>
#include <math.h>

// for simplify definitions
using object::proxy;
using object::base;

namespace object
{
	//
	// stdfunc object
	class stdfunc_call : public base
	{
	public:

		stdfunc_call(int code, size_t argc, const proxy* argv);

		virtual proxy evaluate(bool separable) const;
		virtual std::string toString() const;

	private:

		int m_code;
		myvector<proxy> m_args;
	};
}

namespace 
{
	// Common
	proxy f_equivalence(size_t argc, const proxy* args);

	// Math
	proxy f_plus(size_t argc, const proxy* args);
	proxy f_minus(size_t argc, const proxy* args);
	proxy f_multiplication(size_t argc, const proxy* args);
	proxy f_division(size_t argc, const proxy* args);
	proxy f_sine(size_t argc, const proxy* args);
	proxy f_cosine(size_t argc, const proxy* args);
	proxy f_tangent(size_t argc, const proxy* args);
	proxy f_arcsine(size_t argc, const proxy* args);
	proxy f_arccosine(size_t argc, const proxy* args);
	proxy f_arctangent(size_t argc, const proxy* args);
	proxy f_exponent(size_t argc, const proxy* args);
	proxy f_logarithm(size_t argc, const proxy* args);

	// Lists
	proxy f_car(size_t argc, const proxy* args);
	proxy f_cdr(size_t argc, const proxy* args);
	proxy f_cons(size_t argc, const proxy* args);
	proxy f_list(size_t argc, const proxy* args);
	proxy f_length(size_t argc, const proxy* args);
	proxy f_map(size_t argc, const proxy* args, bool separable);

	// is type
	proxy f_isNull(size_t argc, const proxy* args);
	proxy f_isReal(size_t argc, const proxy* args);
	proxy f_isPair(size_t argc, const proxy* args);
}
///////////////////////////////////////
//
// Public functions
//
///////////////////////////////////////

proxy object::stdfunc::create(binary::link l)
{
	if (l.l.table != binary::link::standard)
	{
		throw std::runtime_error("object::stdfunc::create: it is not standard function");
	}

	return proxy(new stdfunc(l.l.item));
}

// binary::types::lambda - for other code stdfunc looks like lambda
object::stdfunc::stdfunc(int code) : base(object::types::function), m_code(code)
{
}

proxy object::stdfunc::apply(bool separable, size_t argc, const proxy* args) const
{
	proxy p(new stdfunc_call(m_code, argc, args), separable);
	return p; //.simplify();
}

std::string object::stdfunc::toString() const
{
	char buff[20] = "#stdfunc:";

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	_itoa_s(m_code, buff + strlen(buff), sizeof(buff) - strlen(buff), 10);
#else 
	//itoa(m_code, buff + strlen(buff), 10);
#endif

	return buff;
}

///////////////////////////////////////
//
// stdfunc_call functions
//
///////////////////////////////////////

object::stdfunc_call::stdfunc_call(int code, size_t argc, const proxy* argv) 
			: base(object::types::other)
{
	m_simple = false;
	m_code = code;

	if (argc)
	{
		m_args.assign(argv, argv + argc);
	}
}

proxy object::stdfunc_call::evaluate(bool separable) const
{
	switch (m_code)
	{
	case binary::stdfunc::equivalence: return f_equivalence(m_args.size(), m_args.get());

	case binary::stdfunc::plus: return f_plus(m_args.size(), m_args.get());
	case binary::stdfunc::minus: return f_minus(m_args.size(), m_args.get());
	case binary::stdfunc::multiplication: return f_multiplication(m_args.size(), m_args.get());
	case binary::stdfunc::division: return f_division(m_args.size(), m_args.get());
	case binary::stdfunc::sine: return f_sine(m_args.size(), m_args.get());
	case binary::stdfunc::cosine: return f_cosine(m_args.size(), m_args.get());
	case binary::stdfunc::tangent: return f_tangent(m_args.size(), m_args.get());
	case binary::stdfunc::arcsine: return f_arcsine(m_args.size(), m_args.get());
	case binary::stdfunc::arccosine: return f_arccosine(m_args.size(), m_args.get());
	case binary::stdfunc::arctangent: return f_arctangent(m_args.size(), m_args.get());
	case binary::stdfunc::exponent: return f_exponent(m_args.size(), m_args.get());
	case binary::stdfunc::logarithm: return f_logarithm(m_args.size(), m_args.get());
	
	case binary::stdfunc::car: return f_car(m_args.size(), m_args.get());
	case binary::stdfunc::cdr: return f_cdr(m_args.size(), m_args.get());
	case binary::stdfunc::cons: return f_cons(m_args.size(), m_args.get());
	case binary::stdfunc::list: return f_list(m_args.size(), m_args.get());
	case binary::stdfunc::length: return f_length(m_args.size(), m_args.get());
	case binary::stdfunc::map: return f_map(m_args.size(), m_args.get(), separable);

	case binary::stdfunc::isNull: return f_isNull(m_args.size(), m_args.get());
	case binary::stdfunc::isReal: return f_isReal(m_args.size(), m_args.get());
	case binary::stdfunc::isPair: return f_isPair(m_args.size(), m_args.get());
	}

	throw std::runtime_error("object::stdfunc: unknown standard function code");			
}

std::string object::stdfunc_call::toString() const
{
	char buff[20] = "#stdfunc:";

#if defined(_MSC_VER) && (_MSC_VER >= 1400)
	_itoa_s(m_code, buff + strlen(buff), sizeof(buff) - strlen(buff), 10);
#else 
	//itoa(m_code, buff + strlen(buff), 10);
#endif

	return buff;
}

///////////////////////////////////////
//
// Private functions
//
///////////////////////////////////////

namespace
{
	/////////////////////////
	//
	// Equivalence
	//
	/////////////////////////

	// =, eqv?
	proxy f_equivalence(size_t argc, const proxy* args)
	{
		if (argc != 2)
		{
			throw std::runtime_error("stdfunc::equivalence: need two arguments");
		}

		proxy p0 = args[0].simplify();
		proxy p1 = args[1].simplify();

		return object::boolean::create(p0->equiv(p1.get()));
	}

	///////////////////
	//
	// Math
	//
	///////////////////

	// any math function with one argument
	typedef double (* math_one_func) (double arg);
	// any math function
	typedef double (* math_func) (size_t argc, const double* args);

	// Wrapper around any math function with one argument. 
	// Convert argument and return value
	proxy math_wrapper_one(math_one_func f, size_t argc, const proxy* args, const char* name)
	{
		if (argc != 1)
		{
			std::string str;
			str = "stdfunc::";
			str += name;
			str += ": need one argument";
			throw std::runtime_error(str.c_str());
		}

		// prepare argument
		proxy p = args->simplify();

		if (!p->isNumeric())
		{
			throw std::runtime_error("stdfunc::math: need numeric arguments");
		}

		// get object
		const object::numeric* obj = p.get<object::numeric, object::types::numeric>();

		// call function
		double res = f(obj->value());

		// return result
		return object::numeric::create(res);
	}

	// Wrapper around any math function. Convert arguments and return value
	proxy math_wrapper(math_func f, size_t argc, const proxy* args)
	{
		// prepare arguments
		myvector<double> double_args;
		double_args.reserve(argc);

		for (size_t i = 0; i < argc; i++)
		{
			proxy p = args[i].simplify();

			if (!p->isNumeric())
			{
				throw std::runtime_error("stdfunc::math: need numeric arguments");
			}

			const object::numeric* obj = p.get<object::numeric, object::types::numeric>();
			double_args.push_back(obj->value());
		}

		// call function
		double res = f(double_args.size(), double_args.get());

		// return result
		return object::numeric::create(res);
	}

	//
	// Plus
	double math_plus(size_t argc, const double* args)
	{
		double res = 0;

		for (size_t i = 0; i < argc; i++)
		{
			res += args[i];
		}

		return res;
	}

	//
	// Minus
	double math_minus(size_t argc, const double* args)
	{
		if (argc == 1)
		{
			return -args[0];
		}
		else if (argc == 2)
		{
			return args[0] - args[1];
		}
		else
		{
			throw std::runtime_error("stdfunc::minus: need one or two arguments");
		}
	}

	//
	// multiplication
	double math_multiplication(size_t argc, const double* args)
	{
		double res = 1;

		for (size_t i = 0; i < argc; i++)
		{
			res *= args[i];
		}

		return res;
	}

	//
	// divizion
	double math_division(size_t argc, const double* args)
	{
		if (argc == 1)
		{
			if (args[0] == 0)
			{
				throw std::runtime_error("stdfunc::division: division by zero");
			}

			return 1 / args[0];
		}
		else if (argc == 2)
		{
			if (args[1] == 0)
			{
				throw std::runtime_error("stdfunc::division: division by zero");
			}

			return args[0] / args[1];
		}
		else
		{
			throw std::runtime_error("stdfunc::division: need one or two arguments");
		}
	}

	//
	// Plus
	proxy f_plus(size_t argc, const proxy* args)
	{
		return math_wrapper(math_plus, argc, args);
	}

	//
	// Minus
	proxy f_minus(size_t argc, const proxy* args)
	{
		return math_wrapper(math_minus, argc, args);
	}

	//
	// Multiplication
	proxy f_multiplication(size_t argc, const proxy* args)
	{
		return math_wrapper(math_multiplication, argc, args);
	}

	//
	// Division
	proxy f_division(size_t argc, const proxy* args)
	{
		return math_wrapper(math_division, argc, args);
	}

	//
	// Sine
	proxy f_sine(size_t argc, const proxy* args)
	{
		return math_wrapper_one(sin, argc, args, "sin");
	}

	//
	// Cosine
	proxy f_cosine(size_t argc, const proxy* args)
	{
		return math_wrapper_one(cos, argc, args, "cos");
	}

	//
	// Tangent
	proxy f_tangent(size_t argc, const proxy* args)
	{
		return math_wrapper_one(tan, argc, args, "tan");
	}

	//
	// arcsine
	proxy f_arcsine(size_t argc, const proxy* args)
	{
		return math_wrapper_one(asin, argc, args, "asin");
	}

	//
	// arccosine
	proxy f_arccosine(size_t argc, const proxy* args)
	{
		return math_wrapper_one(acos, argc, args, "acos");
	}

	//
	// arctangent
	proxy f_arctangent(size_t argc, const proxy* args)
	{
		return math_wrapper_one(atan, argc, args, "atan");
	}

	//
	// exponent
	proxy f_exponent(size_t argc, const proxy* args)
	{
		return math_wrapper_one(exp, argc, args, "exp");
	}

	//
	// logarithm
	proxy f_logarithm(size_t argc, const proxy* args)
	{
		return math_wrapper_one(log, argc, args, "log");
	}

	/////////////////////////
	//
	// Lists
	//
	/////////////////////////

	proxy f_car(size_t argc, const proxy* args)
	{
		if (argc != 1) 
		{
			throw std::runtime_error("stdfunc::car: need one argument");
		}

		proxy p = args[0].simplify();

		if (!p->isPair())
		{
			throw std::runtime_error("stdfunc::car: need pair argument");
		}

		return p.get<object::pair, object::types::pair>()->car();
	}

	proxy f_cdr(size_t argc, const proxy* args)
	{
		if (argc != 1) 
		{
			throw std::runtime_error("stdfunc::cdr: need one argument");
		}

		proxy p = args[0].simplify();

		if (!p->isPair())
		{
			throw std::runtime_error("stdfunc::cdr: need pair argument");
		}

		return p.get<object::pair, object::types::pair>()->cdr();
	}

	proxy f_cons(size_t argc, const proxy* args)
	{
		if (argc != 2)
		{
			throw std::runtime_error("stdfunc::cons: need two arguments");
		}

		return object::pair::cons(args[0], args[1]);
	}

	proxy f_list(size_t argc, const proxy* args)
	{
		return object::pair::list(argc, args);
	}

	proxy f_length(size_t argc, const proxy* args)
	{
		if (argc != 1)
		{
			throw std::runtime_error("stdfunc::length: need one argument");
		}

		int len = 0;
		proxy p = args[0];

		for (;; len++)
		{
			p = p.simplify();

			if (p->isNull())
			{
				// reach last element
				break;
			}

			if (!p->isPair())
			{
				throw std::runtime_error("stdfunc::length: need pair or null argument");
			}
			
			p = p.get<object::pair, object::types::pair>()->cdr();			
		}

		return object::numeric::create(len);
	}

	proxy f_map(size_t argc, const proxy* args, bool separable)
	{
		if (argc < 1)
		{
			throw std::runtime_error("stdfunc::map: need one or more arguments");
		}

		if (argc == 1)
		{
			// empty arguments list - empty result
			return object::empty_list::create();
		}

		// get function
		proxy func = args[0].simplify();

		if (func->type() != object::types::function)
		{
			throw std::runtime_error("stdfunc::map: first argument must be a function");
		}

		myvector<proxy> results;
		results.reserve(argc);

		for (size_t i = 1; i < argc; i++)
		{
			// apply function to every argument
			// Make separable proxies!
			results.push_back(func->apply(separable, 1, args + i));
		}

		// return list
		return object::pair::list(results.size(), results.get());		
	}

	///////////////////
	//
	// isType functions
	//
	///////////////////

	proxy f_isNull(size_t argc, const proxy* args)
	{
		if (argc != 1)
		{
			throw std::runtime_error("stdfunc::null?: need one argument");
		}

		proxy p = args[0].simplify();
		return object::boolean::create(p->isNull());
	}

	proxy f_isReal(size_t argc, const proxy* args)
	{
		if (argc != 1)
		{
			throw std::runtime_error("stdfunc::real?: need one argument");
		}

		proxy p = args[0].simplify();
		return object::boolean::create(p->isNumeric());
	}

	proxy f_isPair(size_t argc, const proxy* args)
	{
		if (argc != 1)
		{
			throw std::runtime_error("stdfunc::pair?: need one argument");
		}

		proxy p = args[0].simplify();
		return object::boolean::create(p->isPair());
	}

}